__author__ = 'Martin Humby 2015'


from comp_methods import Edgecomp, Vertexcomp, StableSet
from simple_set import SimpleSet

class Edge(Edgecomp):
    def __init__(self, fromVertex, toVertex, weight = 0):
        self.fromVertex = fromVertex
        self.toVertex = toVertex    # the vertex at the far end of this edge
        self.weight = weight        # the weight, cost, distanced got by traversing


class Vertex(Vertexcomp):
    def __init__(self, name, ID):
        self.ID = ID                # integer identifier mapped to/from name
        self.name = name            # any or multiple types relating to the application
        self.edges = StableSet()    # addition order of edges in the set is maintained
        self.in_edges = StableSet()
        self.distance = 0           # basic attribute e.g distance from nearest town

    def addEdge(self, edge):
        self.edges.add(edge)
        edge.toVertex.in_edges.add(edge)

    def order(self):
        return len(self.edges)

    def in_order(self):
            return len(self.in_edges)

    def out_order(self):
        return len(self.edges)

    def resetDistance(self, distance):
        self.distance = distance


class Graph:
    """
    :param directed = True means edges a -> b, b -> a must be defined when required, otherwise this is automatic
    :param vertexClass provides for using Vertex subclasses with additional attributes and/or methods
    """
    def __init__(self, directed = True, vertexClass = Vertex, edgeClass = Edge):
        self.vertexClass = vertexClass
        self.edgeClass = edgeClass
        self.directed = directed
        self.order = 0      # new graph has no vertices
        self.vmap = {}      # map name -> vertex
        self.vertices = []  # lookup vertex ID -> vertex


    def _add_vertex(self, name):
        """
        helper method: adds a new name mapping to vmap and a vertex to vertices indexed by ID
        :param name:
        :return: the vertex, new or existing
        """
        if not name in self.vmap.keys():
            vertex = self.vertexClass(name, self.order)
            self.vmap[name] = vertex
            self.vertices.append(vertex)
            self.order += 1
            return vertex
        else:
            return self.vmap[name]

    def addEdges(self, fromV, *toV):
        """
        If not existing adds the vertex fromV to vmap. Same for toV vertices
        edges are added toV fromV and, if the graph is not directed, toV edges are added with equal traversal cost
        If the first edge is a tuple then all edges must be tuples
        :param fromV: the vertex from which edges run
        :param toV:  can be the vertices by name or tuples (name, edge-weight)
        :param edgeAction: a function giving this method the ability to modify edges
        """

        # add the from vertex to the map and/or get its mapping
        fv = self._add_vertex(fromV)

        # add toV vertices
        aretuple = bool(toV) and isinstance(toV[0], tuple)

        for v in toV:
            if aretuple:
                tv = self._add_vertex(v[0])
                edge = self.edgeClass(fv, tv, v[1])
            else:
                tv = self._add_vertex(v)
                edge = self.edgeClass(fv, tv)
            fv.addEdge(edge)
            if not self.directed:
                tv_edge = self.edgeClass(tv, fv, edge.weight)
                tv.addEdge(tv_edge)

    def vertex(self, name):
        return self.vmap[name]

    def resetDistances(self, distance):
        """
        Sets all vertex distance to a value, generally an infinity or zero.
        Allows reuse of the same graph object with different traversal functions
        :param distance:
        """
        for v in self.vertices:
            v.resetDistance(distance)

    def order(self):
        return self.order

    def __iter__(self):
        for v in self.vertices:
            yield v

    class AdjacencyList:
        def __init__(self, graph):
            self.alist = []
            for v in graph:
                vlist = [v]
                adjacencies = []
                for e in v.edges:
                    adjacencies.append(e)
                vlist.append(adjacencies)
                self.alist.append(vlist)

        def __str__(self):
            s = '[ '
            for v in self.alist:
                s += '[' + str(v[0].name) +'->'
                for e in v[1]:
                    s += '(' + str(e.toVertex.name) + ': ' + str(e.weight)+')'
                s+= '] '
            s += ']'
            return s

    def adjacencies(self):
        return Graph.AdjacencyList(self)
